<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>{% block title %}Hello from Mu! :-){% endblock %}</title>
  <meta name="description" content="A big hello from Mu">
  <meta name="author" content="Add your name here">
  <!-- CSS
  –––––––––––––––––––––––––––––––––––––––––––––––––– -->
  <link rel="stylesheet" href="/static/css/normalize.css">
  <link rel="stylesheet" href="/static/css/skeleton.css">
  <!-- Javascript
  –––––––––––––––––––––––––––––––––––––––––––––––––– -->
  <!-- Mobile Specific Metas (Do not change)
  –––––––––––––––––––––––––––––––––––––––––––––––––– -->
  <meta name="viewport" content="width=device-width, initial-scale=1">
</head>
<body>
  <div class="container">
    {% block content %}{% endblock %}
  </div>
</body>
</html>
