# IMPORTANT
# ---------
# Keep these metadata assignments simple and single-line. They are parsed
# somewhat naively by setup.py and the Windows installer generation script.

__title__ = "mu-editor"
__description__ = "A simple Python editor for beginner programmers."

__version__ = "1.1.0.beta.5"

__license__ = "GPL3"
__url__ = "https://github.com/mu-editor/mu"

__author__ = "Nicholas H.Tollervey"
__email__ = "ntoll@ntoll.org"
